@@ -9,13 +9,16 @@ import threading
 import time
 import os
 import random
+from functools import wraps, partial
 from urllib.parse import urlparse
+from enum import Enum
 import psutil
 
 # asyncio was introduced in version 3.5
 if sys.version_info >= (3, 5):
     try:
         import asyncio_helper
+
         _ASYNCIO_AVAILABLE = True
     except ImportError:
         # This is to deal with python versions below 3.5
@@ -27,21 +30,28 @@ try:
     # This fails when the code is executed directly and not as a part of python package installation,
     # I definitely need a better way to handle this.
     from adbe.adb_helper import (get_adb_shell_property, execute_adb_command2,
-            execute_adb_shell_command, execute_adb_shell_command2,
-            execute_file_related_adb_shell_command, get_package,
-            root_required_to_access_file, get_device_android_api_version, toggle_screen)
-    from adbe.output_helper import print_message, print_error, print_error_and_exit, print_verbose
+                                 execute_adb_shell_command, execute_adb_shell_command2,
+                                 execute_file_related_adb_shell_command, get_package,
+                                 root_required_to_access_file,
+                                 get_device_android_api_version, toggle_screen)
+    from adbe.output_helper import (print_message, print_error, print_error_and_exit,
+                                    print_verbose)
 except ImportError:
     # This works when the code is executed directly.
+    # noinspection PyUnresolvedReferences
     from adb_helper import (get_adb_shell_property, execute_adb_command2,
-            execute_adb_shell_command, execute_adb_shell_command2,
-            execute_file_related_adb_shell_command, get_package,
-            root_required_to_access_file, get_device_android_api_version, toggle_screen)
-    from output_helper import print_message, print_error, print_error_and_exit, print_verbose
+                            execute_adb_shell_command, execute_adb_shell_command2,
+                            execute_file_related_adb_shell_command, get_package,
+                            root_required_to_access_file,
+                            get_device_android_api_version, toggle_screen)
 
+    # noinspection PyUnresolvedReferences
+    from output_helper import (print_message, print_error, print_error_and_exit,
+                               print_verbose)
 
 _KEYCODE_BACK = 4
 _MIN_API_FOR_RUNTIME_PERMISSIONS = 23
+_MIN_API_FOR_DARK_MODE = 29
 
 # Value to be return as 'on' to the user
 _USER_PRINT_VALUE_ON = 'on'
@@ -51,43 +61,22 @@ _USER_PRINT_VALUE_PARTIALLY_ON = 'partially on'
 _USER_PRINT_VALUE_OFF = 'off'
 # Value to be return as 'unknown' to the user
 _USER_PRINT_VALUE_UNKNOWN = 'unknown'
+# Value to be return as 'auto' to the user
+_USER_PRINT_VALUE_AUTO = 'auto'
 
 SCREEN_ON = 1
 SCREEN_OFF = 2
 SCREEN_TOGGLE = 3
 
 
-def _ensure_package_exists(package_name):
-    """
-    Don't call this directly. Instead consider decorating your method with
-    @ensure_package_exists or @ensure_package_exists2
-    :return: True if package_name package is installed on the device
-    """
-    if not _package_exists(package_name):
-        print_error_and_exit("Package %s does not exist" % package_name)
-
-
 # A decorator to ensure package exists
+# Note: This decorator assumes that the decorated func gets package_name as
+# the first parameter
 def ensure_package_exists(func):
-    def func_wrapper(package_name):
-        _ensure_package_exists(package_name)
-        return func(package_name)
-    return func_wrapper
-
-
-# A decorator to ensure package exists with one more argument
-def ensure_package_exists2(func):
-    def func_wrapper(package_name, arg2):
-        _ensure_package_exists(package_name)
-        return func(package_name, arg2)
-    return func_wrapper
-
-
-# A decorator to ensure package exists with two more arguments
-def ensure_package_exists3(func):
-    def func_wrapper(package_name, arg2, arg3):
-        _ensure_package_exists(package_name)
-        return func(package_name, arg2, arg3)
+    def func_wrapper(package_name, *args, **kwargs):
+        if not _package_exists(package_name):
+            print_error_and_exit("Package %s does not exist" % package_name)
+        return func(package_name, *args, **kwargs)
     return func_wrapper
 
 
@@ -97,6 +86,24 @@ def _package_exists(package_name):
     return return_code == 0 and response is not None and len(response.strip()) != 0
 
 
+def print_state_change_decorator(fun, title, get_state_func):
+    # magic sauce to lift the name and doc of the function
+    @wraps(fun)
+    def ret_fun(*args, **kwargs):
+        # Get state before execution
+        current_state = get_state_func()
+        # Call the function
+        returned_value = fun(*args, **kwargs)
+        # Get state after execution
+        # sleep before getting the new value or we might get a stale value in some cases
+        # like mobile-data on/off
+        time.sleep(1)
+        new_state = get_state_func()
+        print_state_change_info(title, current_state, new_state)
+        return returned_value
+    return ret_fun
+
+
 # Source:
 # https://github.com/dhelleberg/android-scripts/blob/master/src/devtools.groovy
 def handle_gfx(value):
@@ -239,6 +246,7 @@ def get_battery_saver_state():
 
 # Source:
 # https://stackoverflow.com/questions/28234502/programmatically-enable-disable-battery-saver-mode
+@partial(print_state_change_decorator, title="Battery saver", get_state_func=get_battery_saver_state)
 def handle_battery_saver(turn_on):
     _error_if_min_version_less_than(19)
     if turn_on:
@@ -361,33 +369,48 @@ def _is_app_running(app_name):
 
 
 def handle_list_devices():
+    device_serials = _get_device_serials()
+
+    if not device_serials:
+        print_error_and_exit('No attached Android device found')
+
+    for device_serial in device_serials:
+        _print_device_info(device_serial)
+
+
+def _get_device_serials() -> [str]:
     cmd = 'devices -l'
     return_code, stdout, stderr = execute_adb_command2(cmd)
     if return_code != 0:
         print_error_and_exit('Failed to execute command %s, error: %s ' % (cmd, stderr))
 
+    device_serials = []
     # Skip the first line, it says "List of devices attached"
     device_infos = stdout.split('\n')[1:]
 
     if len(device_infos) == 0 or (
             len(device_infos) == 1 and len(device_infos[0]) == 0):
-        print_error_and_exit('No attached Android device found')
-    elif len(device_infos) == 1:
-        _print_device_info()
-    else:
-        for device_info in device_infos:
-            if not device_info:
-                continue
-            device_serial = device_info.split()[0]
-            if 'unauthorized' in device_info:
-                device_info = ' '.join(device_info.split()[1:])
-                print_error(
-                    ('Unlock Device "%s" and give USB debugging access to ' +
-                     'this PC/Laptop by unlocking and reconnecting ' +
-                     'the device. More info about this device: "%s"\n') % (
-                         device_serial, device_info))
-            else:
-                _print_device_info(device_serial)
+        return []
+
+    if len(device_infos) == 1:
+        device_serial = device_infos[0].split(" ")[0]
+        device_serials.append(device_serial)
+        return device_serials
+
+    for device_info in device_infos:
+        if not device_info:
+            continue
+        device_serial = device_info.split()[0]
+        if 'unauthorized' in device_info:
+            device_info = ' '.join(device_info.split()[1:])
+            print_error(
+                ('Unlock Device "%s" and give USB debugging access to ' +
+                 'this PC/Laptop by unlocking and reconnecting ' +
+                 'the device. More info about this device: "%s"\n') % (
+                    device_serial, device_info))
+        else:
+            device_serials.append(device_serial)
+    return device_serials
 
 
 def _print_device_info(device_serial=None):
@@ -438,7 +461,7 @@ def _get_top_activity_data():
         app_name, activity_name = regex_result.group(1), regex_result.group(2)
         # If activity name is a short hand then complete it.
         if activity_name.startswith('.'):
-            activity_name = '%s%s' %(app_name, activity_name)
+            activity_name = '%s%s' % (app_name, activity_name)
         return app_name, activity_name
 
     print_error('Unable to extract activity name')
@@ -516,6 +539,7 @@ def get_wifi_state():
     return _USER_PRINT_VALUE_OFF
 
 
+@partial(print_state_change_decorator, title="Wi-Fi", get_state_func=get_wifi_state)
 def set_wifi(turn_on):
     if turn_on:
         cmd = 'svc wifi enable'
@@ -528,6 +552,7 @@ def set_wifi(turn_on):
 
 # Source:
 # https://stackoverflow.com/questions/26539445/the-setmobiledataenabled-method-is-no-longer-callable-as-of-android-l-and-later
+@partial(print_state_change_decorator, title="Mobile data", get_state_func=get_mobile_data_state)
 def handle_mobile_data(turn_on):
     if turn_on:
         cmd = 'svc data enable'
@@ -632,6 +657,7 @@ def get_mobile_data_saver_state():
 
 
 # https://developer.android.com/training/basics/network-ops/data-saver.html
+@partial(print_state_change_decorator, title="Mobile data saver", get_state_func=get_mobile_data_saver_state)
 def handle_mobile_data_saver(turn_on):
     if turn_on:
         cmd = 'cmd netpolicy set restrict-background true'
@@ -662,6 +688,9 @@ def get_dont_keep_activities_in_background_state():
 # adb shell settings put global always_finish_activities true might not work on all Android versions.
 # It was in system (not global before ICS)
 # adb shell service call activity 43 i32 1 followed by that
+@partial(print_state_change_decorator,
+         title="Don\'t keep activities",
+         get_state_func=get_dont_keep_activities_in_background_state)
 def handle_dont_keep_activities_in_background(turn_on):
     # Till Api 25, the value was True/False, above API 25, 1/0 work. Source: manual testing
     use_true_false_as_value = get_device_android_api_version() <= 25
@@ -701,11 +730,15 @@ def get_show_taps_state():
         print_error('Failed to get current state of "show user taps" setting')
         return _USER_PRINT_VALUE_UNKNOWN
 
-    if int(stdout.strip()) == 1:
+    stdout = stdout.strip()
+    if stdout == 'null':
+        return _USER_PRINT_VALUE_OFF
+    if int(stdout) == 1:
         return _USER_PRINT_VALUE_ON
     return _USER_PRINT_VALUE_OFF
 
 
+@partial(print_state_change_decorator, title="Show user taps", get_state_func=get_show_taps_state)
 def toggle_show_taps(turn_on):
     if turn_on:
         value = 1
@@ -732,6 +765,9 @@ def get_stay_awake_while_charging_state():
 
 
 # Source: https://developer.android.com/reference/android/provider/Settings.Global.html#STAY_ON_WHILE_PLUGGED_IN
+@partial(print_state_change_decorator,
+         title="Stay awake while charging",
+         get_state_func=get_stay_awake_while_charging_state)
 def stay_awake_while_charging(turn_on):
     if turn_on:
         # 1 for USB charging, 2 for AC charging, 4 for wireless charging. Or them together to get 7.
@@ -898,7 +934,7 @@ def get_permissions_in_permission_group(permission_group):
     return None
 
 
-@ensure_package_exists3
+@ensure_package_exists
 def grant_or_revoke_runtime_permissions(package_name, action_grant, permissions):
     _error_if_min_version_less_than(23)
     if action_grant:
@@ -921,22 +957,95 @@ def _get_all_packages(pm_cmd):
     return packages
 
 
-def list_all_apps():
-    cmd = 'pm list packages'
-    packages = _get_all_packages(cmd)
-    print('\n'.join(packages))
+# "dumpsys package" is more accurate than "pm list packages" but that means the results of
+# list_all_apps are now different from list_system_apps, print_list_non_system_apps, and
+# list_debug_apps
+# For now, we can live with this discrepancy but in the longer run we want to fix those
+# other functions as well
+# https://stackoverflow.com/questions/63416599/adb-shell-pm-list-packages-missing-some-packages
+def get_list_all_apps():
+    """This function return a list of installed applications, error message and command
+    execution error
+    :returns: tuple(all_apps, err_msg, error)
+        WHERE
+        list[str] all_apps is a string list of all installed packages
+        str err_msg is the error message to display
+        str error is the command execution error message
+    :Example:
+    >>> import adbe.adb_enhanced as adb_e
+    >>> import adbe.adb_helper as adb_h
+    >>> adb_h.set_device_id("emulator-5554")
+    >>> list_apps, err_msg, err = adb_e.get_list_all_apps()
+    """
+    # https://developer.android.com/studio/command-line/dumpsys
+    cmd = 'dumpsys package'
+    pattern_packages = re.compile('Package \\[(.*?)\\]')
+    return_code, result, err = execute_adb_shell_command2(cmd)
+    if return_code != 0:
+        err_msg = 'Command "%s" failed, something is wrong' % cmd
+        return None, err_msg, err
+    all_apps = re.findall(pattern_packages, result)
+    # Get the unique results
+    all_apps = sorted(list(dict.fromkeys(all_apps)))
+    return all_apps, None, None
+
+
+def print_list_all_apps():
+    """This function print list of all installed packages or error message if an error
+    occurred
+    :returns: None
+    """
+    all_apps, err_msg, err = get_list_all_apps()
+    if err:
+        print_error_and_exit(err_msg)
+        return
+    print_message('\n'.join(all_apps))
+
+
+def get_list_system_apps():
+    """This function return a list of installed system applications
+    :returns: system_apps_packages
+        WHERE
+        list[str] system_apps_packages is a strings list of installed system packages
+    :Example:
+    >>> import adbe.adb_enhanced as adb_e
+    >>> import adbe.adb_helper as adb_h
+    >>> adb_h.set_device_id("DEVICE_ID")
+    >>> list_sys_apps = adb_e.get_list_system_apps()
+    """
+    cmd = 'pm list packages -s'
+    system_apps_packages = _get_all_packages(cmd)
+    return system_apps_packages
 
 
 def list_system_apps():
-    cmd = 'pm list packages -s'
-    packages = _get_all_packages(cmd)
+    """This function print list of installed system packages
+    :returns: None
+    """
+    packages = get_list_system_apps()
     print('\n'.join(packages))
 
 
-def list_non_system_apps():
+def get_list_non_system_apps():
+    """Return a list of installed third party applications.
+    :returns: third_party_pkgs
+        WHERE
+        list[str] third_party_pkgs is a strings list of installed third party packages
+    :Example:
+    >>> import adbe.adb_enhanced as adb_e
+    >>> import adbe.adb_helper as adb_h
+    >>> adb_h.set_device_id("DEVICE_ID")
+    >>> list_sys_apps = adb_e.get_list_non_system_apps()
+    """
     cmd = 'pm list packages -3'
-    packages = _get_all_packages(cmd)
-    print('\n'.join(packages))
+    return _get_all_packages(cmd)
+
+
+def print_list_non_system_apps():
+    """Print list of installed third party packages.
+    :returns: None
+    """
+    print('\n'.join(get_list_non_system_apps()))
 
 
 def list_debug_apps():
@@ -979,7 +1088,7 @@ def _is_debug_package(app_name):
     app_info_dump = execute_adb_shell_command(pm_cmd, piped_into_cmd=grep_cmd)
     if app_info_dump is None or app_info_dump.strip() == '0':
         return app_name, False
-    elif app_info_dump.strip() == '1' or app_info_dump.strip() == '2':
+    elif app_info_dump.strip() in ('1', '2'):
         return app_name, True
     else:
         print_error_and_exit('Unexpected output for %s | %s = %s' % (pm_cmd, grep_cmd, app_info_dump))
@@ -1026,7 +1135,7 @@ def _is_allow_backup_package(app_name):
     app_info_dump = execute_adb_shell_command(pm_cmd, piped_into_cmd=grep_cmd)
     if app_info_dump is None or app_info_dump.strip() == '0':
         return app_name, False
-    elif app_info_dump.strip() == '1' or app_info_dump.strip() == '2':
+    elif app_info_dump.strip() in ('1', '2'):
         return app_name, True
     else:
         print_error_and_exit('Unexpected output for %s | %s = %s' % (pm_cmd, grep_cmd, app_info_dump))
@@ -1050,12 +1159,12 @@ def get_standby_bucket(package_name):
     result = execute_adb_shell_command(cmd)
     if result is None:
         print_error_and_exit(_USER_PRINT_VALUE_UNKNOWN)
-    print_verbose('App standby bucket for \"%s\" is %s' %(
+    print_verbose('App standby bucket for \"%s\" is %s' % (
         package_name, _APP_STANDBY_BUCKETS.get(int(result), _USER_PRINT_VALUE_UNKNOWN)))
     print(_APP_STANDBY_BUCKETS.get(int(result), _USER_PRINT_VALUE_UNKNOWN))
 
 
-@ensure_package_exists2
+@ensure_package_exists
 def set_standby_bucket(package_name, mode):
     _error_if_min_version_less_than(28)
     cmd = 'am set-standby-bucket %s %s' % (package_name, mode)
@@ -1078,7 +1187,7 @@ def calculate_standby_mode(args):
 
 
 # Source: https://developer.android.com/preview/features/power
-@ensure_package_exists2
+@ensure_package_exists
 def apply_or_remove_background_restriction(package_name, set_restriction):
     _error_if_min_version_less_than(28)
     appops_cmd = 'cmd appops set %s RUN_ANY_IN_BACKGROUND %s' % (
@@ -1174,7 +1283,7 @@ def pull_file(remote_file_path, local_file_path, copy_ancillary=False):
             if not _file_exists(tmp_db_file):
                 continue
             if copy_ancillary:
-                pull_file(tmp_db_file, '%s-%s' %(local_file_path, suffix), copy_ancillary=True)
+                pull_file(tmp_db_file, '%s-%s' % (local_file_path, suffix), copy_ancillary=True)
             else:
                 print_error('File \"%s\" has an ancillary file \"%s\" which should be copied.\nSee %s for details'
                             % (remote_file_path, tmp_db_file,
@@ -1338,9 +1447,9 @@ def _get_permissions_info_above_api_23(app_info_dump):
             runtime_denied_permissions.append(permission)
     runtime_not_granted_permissions = list(filter(
         lambda p: p not in runtime_granted_permissions and
-        p not in runtime_denied_permissions and
-        p not in install_time_granted_permissions and
-        p not in install_time_denied_permissions, requested_permissions))
+                  p not in runtime_denied_permissions and
+                  p not in install_time_granted_permissions and
+                  p not in install_time_denied_permissions, requested_permissions))
 
     permissions_info_msg = ''
     permissions_info_msg += '\nPermissions:\n\n'
@@ -1406,12 +1515,13 @@ def print_app_signature(app_name):
 
 
 # Uses abe.jar taken from https://sourceforge.net/projects/adbextractor/
-@ensure_package_exists2
+@ensure_package_exists
 def perform_app_backup(app_name, backup_tar_file):
     # TODO: Add a check to ensure that the screen is unlocked
     password = '00'
     print_verbose('Performing backup to backup.ab file')
-    print_message('you might have to confirm the backup manually on your device\'s screen, enter \"%s\" as password...' % password)
+    print_message(
+        'you might have to confirm the backup manually on your device\'s screen, enter \"%s\" as password...' % password)
 
     def backup_func():
         # Create backup.ab
@@ -1470,9 +1580,24 @@ def perform_install(file_path):
 
 
 @ensure_package_exists
-def perform_uninstall(app_name):
+def perform_uninstall(app_name, first_user):
     print_verbose('Uninstalling %s' % app_name)
-    return_code, _, stderr = execute_adb_command2('uninstall %s' % app_name)
+    cmd = ""
+    if first_user:
+        # For system apps, that cannot uninstalled,
+        # this command uninstalls them for user 0 without doing a system uninstall
+        # since that would fail.
+        # https://www.xda-developers.com/uninstall-carrier-oem-bloatware-without-root-access/
+        cmd = "--user 0"
+    return_code, _, stderr = execute_adb_shell_command2('pm uninstall %s %s' % (cmd, app_name))
+    if return_code == 0:
+        return
+
+    if not cmd:
+        print_message("Uninstall failed, trying to uninstall for user 0...")
+        cmd = "--user 0"
+        return_code, _, stderr = execute_adb_shell_command2('pm uninstall %s %s' % (cmd, app_name))
+
     if return_code != 0:
         print_error('Failed to uninstall %s, stderr: %s' % (app_name, stderr))
 
@@ -1537,6 +1662,65 @@ def _is_emulator():
     return qemu is not None and qemu.strip() == '1'
 
 
+def enable_wireless_debug():
+    code, result, stderr = execute_adb_shell_command2("ip address")
+    if code != 0:
+        print_error_and_exit('Failed to switch device to wireless debug mode, stderr: '
+                             '%s' % stderr)
+
+    # Check, that phone connected to wlan
+    matching = re.findall(r"inet ([\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}).*wlan0$",
+                          result, re.MULTILINE)
+    if matching is None or not matching:
+        print_error_and_exit('Failed to switch device to wireless debug mode')
+
+    ip = matching[0]
+
+    code, _, stderr = execute_adb_command2("tcpip 5555")
+    if code != 0:
+        print_error_and_exit('Failed to switch device %s to wireless debug mode, '
+                             'stderr: %s' % (ip, stderr))
+
+    code, _, stderr = execute_adb_command2("connect %s" % ip)
+    if code != 0:
+        print_error_and_exit('Cannot enable wireless debugging. Error: %s' % stderr)
+        return False
+    print_message('Connected via IP now you can disconnect the cable\nIP: %s' % ip)
+    return True
+
+
+def disable_wireless_debug():
+    device_serials = _get_device_serials()
+
+    if not device_serials:
+        print_error_and_exit('No connected device found')
+        return
+
+    ip_list = []
+    for device_serial in device_serials:
+        ips = re.findall(r"([\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}:[\d]{1,5})", device_serial, 0)
+        if not ips:
+            print_verbose('Not a IP connect device, serial: %s' % device_serial)
+            continue
+        if len(ips) > 1:
+            print_error('Malformed device IP: %s' % device_serial)
+        print_verbose('Found an IP connected ADB session: %s' % ips[0])
+        ip_list.append(ips[0])
+
+    result = True
+
+    for ip in ip_list:
+        code, _, stderr = execute_adb_command2('disconnect %s' % ip)
+        if code != 0:
+            print_error('Failed to disconnect %s: %s' % (ip, stderr))
+            result = False
+        else:
+            print_message('Disconnected %s' % ip)
+
+    if not result:
+        print_error_and_exit('')
+
+
 def switch_screen(switch_type):
     if switch_type == SCREEN_TOGGLE:
         c, o, e = toggle_screen()
@@ -1564,6 +1748,43 @@ def switch_screen(switch_type):
     return o
 
 
+def get_dark_mode() -> str:
+    _error_if_min_version_less_than(_MIN_API_FOR_DARK_MODE)
+    return_code, stdout, stderr = execute_adb_shell_settings_command2('get secure ui_night_mode')
+    if return_code != 0:
+        print_error('Failed to get current UI mode: %s' % stderr)
+        return _USER_PRINT_VALUE_UNKNOWN
+    if stdout == 'null':
+        return _USER_PRINT_VALUE_UNKNOWN
+    val = int(stdout)
+    if val == 2:
+        return _USER_PRINT_VALUE_ON
+    elif val == 1:
+        return _USER_PRINT_VALUE_OFF
+    elif val == 0:
+        return _USER_PRINT_VALUE_AUTO
+    else:
+        return 'Unknown: %d' % val
+
+
+# This code worked for emulator on API 29.
+# It didn't work for unrooted device on API 30.
+# I am not sure if the problem is rooting or API version
+def set_dark_mode(force: bool) -> None:
+    """
+    :param force: if true, force dark mode, if false don't
+    """
+    _error_if_min_version_less_than(_MIN_API_FOR_DARK_MODE)
+
+    if force:
+        execute_adb_shell_command2('service call uimode 4 i32 2')
+        # There are reports of the following command, it didn't work for me
+        # even on a rooted device when ran as a super-user
+        # execute_adb_shell_command2('setprop persist.hwui.force_dark true')
+    else:
+        execute_adb_shell_command2('service call uimode 4 i32 1')
+
+
 def print_notifications():
     # Noredact is neeed on >= Android 6.0 to see title and text
     code, output, err = execute_adb_shell_command2("dumpsys notification --noredact")
@@ -1571,11 +1792,11 @@ def print_notifications():
         print_error_and_exit("Something gone wrong on "
                              "fetching notification info. Error: %s" % err)
     notification_records = re.findall(r"\s*NotificationRecord\(.*", output, re.MULTILINE)
-    for i in range(0, len(notification_records)):
-        output_for_this_notification = output.split(notification_records[i])[1]
+    for i, notification_record in enumerate(notification_records):
+        output_for_this_notification = output.split(notification_record)[1]
         if i + 1 < len(notification_records):
-            output_for_this_notification = output_for_this_notification.split(notification_records[i+1])[0]
-        notification_package = re.findall(r"pkg=(\S*)", notification_records[i])[0]
+            output_for_this_notification = output_for_this_notification.split(notification_records[i + 1])[0]
+        notification_package = re.findall(r"pkg=(\S*)", notification_record)[0]
         titles = re.findall("android.title=(.*)", output_for_this_notification)
         if len(titles) > 0:
             notification_title = titles[0]
@@ -1588,11 +1809,12 @@ def print_notifications():
             notification_text = None
         notification_actions = []
         action_strings = re.findall(r"actions=\{(.*?)\n\}", output_for_this_notification, re.MULTILINE | re.DOTALL)
-        if len(action_strings) > 0:
-            if i+1 >= len(notification_records) or \
-                    output_for_this_notification.find(action_strings[0]) > output_for_this_notification.find(notification_records[i+1]):
-                for actions in action_strings[0].split('\n'):
-                    notification_actions += re.findall(r"\".*?\"", actions)
+        if len(action_strings) > 0 and (i + 1 >= len(notification_records) or
+                                        output_for_this_notification.find(
+                                            action_strings[0]) > output_for_this_notification.find(
+                    notification_records[i + 1])):
+            for actions in action_strings[0].split('\n'):
+                notification_actions += re.findall(r"\".*?\"", actions)
 
         print_message('Package: %s' % notification_package)
         if notification_title:
@@ -1602,3 +1824,162 @@ def print_notifications():
         for action in notification_actions:
             print_message('Action: %s' % action)
         print_message('')
+
+
+# Alarm Enum
+class AlarmEnum(Enum):
+    TOP = 't'
+    PENDING = 'p'
+    HISTORY = 'h'
+    ALL = 'a'
+
+
+def print_history_alarms(output_dump_alarm, padding):
+    print("App Alarm history")
+
+    pattern_pending_alarm = \
+        re.compile(r'(?<=App Alarm history:)'
+                   r'.*?(?=Past-due non-wakeup alarms)',
+                   re.DOTALL)
+    alarm_to_parse = re.sub(r' +', ' ',
+                            re.search(pattern_pending_alarm, output_dump_alarm).
+                            group(0)).split("\n")[1:-1]
+
+    for line in alarm_to_parse:
+        package_name = line[0:line.find(",")]
+        # +1 to escape ',' before user id
+        fields = line[line.find(",") + 1:].split(":")
+        user_id = fields[0]
+        print("%sPackage name: %s" % (padding, package_name))
+        print("%sUser ID: %s" % (padding * 2, user_id))
+        # History might be missing for new alarms
+        if len(fields) >= 2:
+            history = fields[1]
+            print("%shistory: %s" % (padding * 2, history))
+
+
+def print_top_alarms(output_dump_alarm, padding):
+    print("Top Alarms:")
+    pattern_top_alarm = re.compile(r'(?<=Top Alarms:\n).*?(?=Alarm Stats:)',
+                                   re.DOTALL)
+    alarm_to_parse = re.sub(
+        r' +', ' ',
+        re.search(pattern_top_alarm, output_dump_alarm).group(0)).split("\n")
+    temp_dict = {}
+    for i, alarm_i in enumerate(alarm_to_parse):
+        if re.match(r"^\+", alarm_i):
+            temp_dict[alarm_i] = alarm_to_parse[i + 1]
+            i += 1
+
+    for key, value in temp_dict.items():
+        # key example: +2m19s468ms running, 0 wakeups, 708 alarms: 1000:android
+        # value example: *alarm*:com.android.server.action.NETWORK_STATS_POLL
+        temp = key.split(',')
+        running_time = temp[0].split(" ")[0]
+        nb_woke_up = temp[1].strip().split(" ")[0]
+        nb_alarms = temp[2].strip().split(" ")[0]
+        uid = temp[2].strip().split(":")[1].strip()
+        package_name = temp[2].strip().split(":")[2].strip()
+        action = value.split(":")[1]
+        print("%sPackage name: %s" % (padding, package_name))
+        print("%sAction: %s" % (padding * 2, action))
+        print("%sRunning time: %s" % (padding * 2, running_time))
+        print("%sNumber of device woke up: %s" % (padding * 2, nb_woke_up))
+        print("%sNumber of alarms: %s" % (padding * 2, nb_alarms))
+        print("%sUser ID: %s" % (padding * 2, uid))
+
+
+def print_pending_alarms(output_dump_alarm, padding):
+    print("Pending Alarms:")
+    pattern_pending_alarm = \
+        re.compile(
+            r'(?<=Pending alarm batches:)'
+            r'.*?(?=(Pending user blocked background alarms|Past-due non-wakeup alarms))',
+            re.DOTALL)
+    alarm_to_parse = re.sub(
+        r' +', ' ',
+        re.search(pattern_pending_alarm, output_dump_alarm).group(0)).split("\n")[1:-1]
+    for line in alarm_to_parse:
+        line = line.strip()
+        if not line.startswith("Batch"):
+            continue
+
+        pattern_batch_info = re.compile(r'(?<=Batch\{).*?(?=\}:)',
+                                        re.DOTALL)
+        info = re.search(pattern_batch_info, line).group(0).split(" ")
+        print("%sID: %s" % (padding, info[0]))
+        print("%sNumber of alarms: %s" % (padding * 2, info[1].split("=")[1]))
+        print_verbose("%sStart: %s" % (padding * 2, info[2].split("=")[1]))
+        print_verbose("%sEnd: %s" % (padding * 2, info[3].split("=")[1]))
+        if "flgs" in line:
+            # TO-DO: translate the flags
+            print_verbose("%sflag: %s" % (padding * 2, info[4].split("=")[1]))
+
+        if line.startswith("RTC") or line.startswith("RTC_WAKEUP") or \
+                line.startswith("ELAPSED") or line.startswith("ELAPSED_WAKEUP"):
+            pattern_between_brackets = re.compile(r'(?<=\{).*?(?=\})',
+                                                  re.DOTALL)
+            info = re.search(pattern_between_brackets, line).group(0).split(" ")
+            print("%sAlarm #%s:" % (padding * 2, line.split("#")[1].split(":")[0]))
+            print_verbose("%sType: %s" % (padding * 2, line.split("#")[0]))
+            print_verbose("%sID: %s" % (padding * 2, info[0]))
+            print_verbose("%sType: %s" % (padding * 2, info[2]))
+            print_verbose("%sWhen: %s" % (padding * 2, info[4]))
+            print("%sPackage: %s" % (padding * 2, info[5]))
+
+
+def alarm_manager(param):
+    cmd = "dumpsys alarm"
+    api_version = get_device_android_api_version()
+    err_msg_api = "Your Android version (API 28 and bellow) does not support " \
+                  "listing pending alarm"
+
+    c, o, e = execute_adb_shell_command2(cmd)
+    if c != 0:
+        print_error_and_exit("Something gone wrong on "
+                             "dumping alarms. Error: %s" % e)
+        return
+
+    if not isinstance(param, AlarmEnum):
+        print_error("Not supported parameter")
+        return
+
+    run_all = 0
+    padding = ""
+    if param == AlarmEnum.ALL:
+        run_all = 1
+        padding = "\t"
+
+    if param == AlarmEnum.TOP or run_all == 1:
+        print_top_alarms(o, padding)
+
+    if param == AlarmEnum.PENDING or run_all == 1:
+        if api_version > 28:
+            print_pending_alarms(o, padding)
+        else:
+            print_error(err_msg_api)
+
+    if param == AlarmEnum.HISTORY or run_all == 1:
+        if api_version > 28:
+            print_history_alarms(o, padding)
+        else:
+            print_error(err_msg_api)
+
+
+# This permissions group seems to have been removed in API 29 and beyond.
+# https://github.com/ashishb/adb-enhanced/runs/1799363523?check_suite_focus=true
+def is_permission_group_unavailable_after_api_29(permission_group):
+    return permission_group in [
+        'android.permission-group.CONTACTS',
+        'android.permission-group.MICROPHONE',
+        'android.permission-group.LOCATION',
+        'android.permission-group.SMS',
+    ]
+
+
+def print_state_change_info(state_name, old_state, new_state):
+    if old_state != new_state:
+        print_message('"%s" state changed from "%s" -> "%s"' % (
+            state_name, old_state, new_state))
+    else:
+        print_message('"%s" state unchanged (%s)' % (state_name, old_state))
